from __future__ import print_function
import numpy as np
from readRanges import readRanges
from readCoeffs import readCoeffs

ranges = readRanges('ranges.txt')
muTerms = readCoeffs('muCoefficients.txt', ranges)
sigmaTerms = readCoeffs('sigmaCoefficients.txt', ranges)

def chlModel(modisData):
   '''calculates pixel bias and sd as a function of input variables'''
   
   if muTerms['parameter'] != 'Mu':
      raise NameError('Unexpected mu parameter', muTerms['parameter'])
   if sigmaTerms['parameter'] != 'Sigma':
      raise NameError('Unexpected sigma parameter', sigmaTerms['parameter'])
   l2flags = modisData['l2flags']
   flags = {}
   for i in range(32):
      flags['flag' + str(i + 1)] = (l2flags & 2**i) != 0 # L3
   modisData['chlL3masked'] = modisData['chl'].copy()
   l3flags = [4, 6, 9, 11, 13, 15, 17, 20, 22, 23, 26, 27]
   for i in l3flags:
      modisData['chlL3masked'].mask |= flags['flag' + str(i)]
   bias = np.ma.zeros(l2flags.shape)
   bias.mask = modisData['chl'].mask
   for term in muTerms:
      if term in ['linkFunction', 'parameter', 'intercept']:
         continue
      if term[:4] == 'flag':
         data = flags[term]
      else:
         if not term in modisData:
            print('mu term', term, 'not found in modisData')
            continue
         data = modisData[term]
      if 'mask' in data:
         bias.mask[data.mask] = True
      item = muTerms[term]
      if isinstance(item, dict):
         myRange = item['range']
         coeffs = item['coeffs']
         scope = data < myRange[0]
         bias[scope] += np.polyval(coeffs[0], data[scope])
         scope = np.logical_and(data >= myRange[0], data <= myRange[1])
         bias[scope] += np.polyval(coeffs[1], data[scope])
         scope = data > myRange[1]
         bias[scope] += np.polyval(coeffs[2], data[scope])
      else:
         bias += item * data
   bias += muTerms['intercept']
   if muTerms['linkFunction'] == 'log':
      print('WARNING: using log link for mu')
      bias = np.exp(bias)
   sd = np.ma.zeros(l2flags.shape)
   sd.mask = modisData['chl'].mask
   for term in sigmaTerms:
      if term in ['linkFunction', 'parameter', 'intercept']:
         continue
      if term[:4] == 'flag':
         data = flags[term]
      else:
         if not term in modisData:
            print('sigma term', term, 'not found in modisData')
            continue
         data = modisData[term]
      if 'mask' in data:
         sd.mask[data.mask] = True
      item = sigmaTerms[term]
      if isinstance(item, dict):
         myRange = item['range']
         coeffs = item['coeffs']
         scope = data < myRange[0]
         sd[scope] += np.polyval(coeffs[0], data[scope])
         scope = np.logical_and(data >= myRange[0], data <= myRange[1])
         sd[scope] += np.polyval(coeffs[1], data[scope])
         scope = data > myRange[1]
         sd[scope] += np.polyval(coeffs[2], data[scope])
      else:
         sd += item * data
   sd += sigmaTerms['intercept']
   if sigmaTerms['linkFunction'] == 'log': # normal
      sd = np.exp(sd)
   return bias, sd

