from __future__ import print_function
import numpy as np,csv,png,glob,scipy,itertools
from netCDF4 import Dataset
from scipy import misc
from chlModel import chlModel
from datetime import date, timedelta
from matplotlib import pyplot as plt
import rpy2.robjects as robjects
from scipy.interpolate import interp1d
from scipy.stats import chi2
from pyresample import kd_tree, geometry
from PIL import Image, ImageFont, ImageDraw 

processing = True

def extrap1d(interpolator): # https://stackoverflow.com/questions/2745329/how-to-make-scipy-interpolate-give-an-extrapolated-result-beyond-the-input-range
  xs = interpolator.x
  ys = interpolator.y

  def pointwise(x):
    if x < xs[0]:
      return ys[0]+(x-xs[0])*(ys[1]-ys[0])/(xs[1]-xs[0])
    elif x > xs[-1]:
      return ys[-1]+(x-xs[-1])*(ys[-1]-ys[-2])/(xs[-1]-xs[-2])
    else:
      return interpolator(x)

  def ufunclike(xs):
    return scipy.array(map(pointwise, scipy.array(xs)))

  return ufunclike

root2=2.**.5
ln10=np.log(10)
dn1=1.*.015-2.
dn254=254.*.015-2.
limits=[35,61,-5,20]
resolution=.01
fill=-9.96920996839e+36
nlc=int(round((limits[1]-limits[0])/resolution))
nsc=int(round((limits[3]-limits[2])/resolution))
dtom = 20037508.3 / 180 # degrees to m
area_def = geometry.AreaDefinition('shelf', 'European shelf seas', 'cyl',
  {'a': '6378144.0', 'b': '6356759.0', 'lat_0': '0.00', 'lat_ts': '0.00',
  'lon_0': '0.00', 'proj': 'eqc'}, nsc, nlc,
  [i * dtom for i in [-5., 35., 20., 61.]])

def loadModisData(names):
 for name in names:
  if name[:3]=='Rrs':
   var=ncv['Rrs_'+name[3:]]
  elif name[:4]=='taua':
   var=ncv['aot'+name[4:]]
  else:
   var=ncv[name]
  data=var[::-1,::-1]
  if data.dtype in [np.dtype('>i2'), np.int16]:
   data=np.array(data)
   data=np.ma.masked_array(data,mask=data==-32767)
   data=data*var.slope+var.intercept
  else:
   data=np.ma.masked_array(data,mask=data<-32766)
  modisData[name]=data

def writeChl(chlor,fileName,land=None,log=False,lim=None):
 if lim is not None:
   if lim is True: lim = [1,chlor.max()]
   chl8bit=np.array(np.ma.round(np.ma.clip(
     (chlor-lim[0])*254/(lim[1]-lim[0]),1,254)), dtype=int)
   chl8bit[chlor == 0] = 255
 elif log:
   chl8bit=np.array(np.ma.round((np.ma.clip(chlor/ln10,dn1,dn254)
     + 2)/0.015), dtype=int)
 else:
   chl8bit=np.array(np.ma.round((np.ma.clip(np.ma.log10(np.maximum(chlor,
     .01)),dn1,dn254) + 2)/0.015), dtype=int)
 if type(chlor) is np.ma.MaskedArray:
   chl8bit[chlor.mask]=255
 chl8bit[flags['flag2'] if land is None else land != 0]=0
 with open(fileName,'wb') as f:
  ctPng.write(f,chl8bit)
 return(chl8bit)

def writeSatChl(chlor,error,fileName,land=None):
 chlIndex=np.ma.clip(np.ma.round(
   (np.ma.log10(chlor) + 2)/0.015).astype(int),1,254)-1
 satIndex=np.ma.clip(np.ma.round(
   ((np.ma.exp(error)-1)*100.-p0)*100./p1).astype(int),0,100)
 chlRgb=rgbScale[satIndex,chlIndex,:]
 chlRgb[chlor.mask,:]=255
 for i in range(3):
  chlRgb[flags['flag2'] if land is None else land != 0,i]=chlScale[0][i]
 s=chlRgb.shape
 with open(fileName,'wb') as f:
  rgbPng.write(f,chlRgb.reshape((s[0],s[1]*s[2])))
 return(chlRgb)

def sat(x):
 return(.5*(1.-scipy.special.erf((x-e0)/(root2*sigma))))

ndata=np.zeros((180,360))
with open('matchups2satAgeDayModels.csv') as f:
   reader=csv.reader(f)
   header=next(reader)
   latPos=header.index('latS')
   lonPos=header.index('lonS')
   stationPos=header.index('STATION')
   oldStation=None
   for data in reader:
      station=data[stationPos]
      if station==oldStation:
         continue
      oldStation=station
      lat,lon=(int(round(float(data[latPos]))),int(round(float(data[lonPos]))))
      ndata[lat+90,lon+180]+=1

ndata=np.ma.array(ndata.astype(int),mask=ndata==0)  
with Dataset('ndata.nc','w') as nc:
   d = nc.createDimension('lat', 180)
   d = nc.createDimension('lon', 360)
   v = nc.createVariable('latitude', 'f', ('lat'))
   v[:] = np.arange(180.)-89.5
   v = nc.createVariable('longitude', 'f', ('lon'))
   v[:] = np.arange(360.)-179.5
   var = nc.createVariable('ndata', 'i', ('lat','lon'), fill_value = -999.)
   var[:]=ndata

launchDate = date(2002, 5, 4) # MODIS Aqua
l3flags = [4, 6, 9, 11, 13, 15, 17, 20, 22, 23, 26, 27]
chlScale=[]
with open('chl_scalebar0grey255white.txt') as f:
 for i in range(6): line=f.readline()
 for line in f: chlScale+=[tuple([int(i) for i in line.split()[-3:]])]

rgbScale=np.empty((101,254,3))
sat=np.arange(101.)[::-1]/100
isat=1.-sat
p0 = 0.
rgbScale[0,:,:]=chlScale[1:-1]
rgbScale[1:,:,:]=[[np.array(chlScale[j])*sat[i]+255*isat[i]
  for j in range(1,255)] for i in range(1,101)]
rgbScale=np.round(rgbScale).astype(int)
rgbScale[[33,67],-3:]=0
palettePng=png.Writer(254,12)
satPalette2Png=png.Writer(256,102)
paletteScale=chlScale[1:-1]
paletteScaleTicks=list(paletteScale) # create a copy
img = Image.new('1', (2560, 160), 255)
draw = ImageDraw.Draw(img)
myFont = ImageFont.truetype("Palatino Linotype.ttf", 120)
draw.text((0, 0),'0.01',0,font=myFont)
w, h = draw.textsize('0.01', font=myFont)
wpc, hpc = draw.textsize('300%', font=myFont)
img.close()
img = Image.new('1', (2560+w, 160), 255)
draw = ImageDraw.Draw(img)
draw.text((0, 0),'0.01',0,font=myFont)
for chl2, s in zip([.03,.1,.3,1,3,10,30,60],
  ['0.03','0.1','0.3','1','3','10','30','60']):
 i=int(round((np.log10(chl2) + 2)/0.015))-1
 rgbScale[:3,i,:]=0
 paletteScaleTicks[i]=(0,0,0)
 w2, h = draw.textsize(s, font=myFont)
 i=int((w - w2) * .5 + round((np.log10(chl2) + 2)/0.0015))-10
 draw.text((i, 0),s,0,font=myFont)

img.save('paletteNumbers.png')
p = misc.imread('paletteNumbers.png')
while p[:,-1].min():
  p = p[:,:-1]

misc.imsave('paletteNumbers.png',p)
paletteScaleTicks2=list(paletteScale) # create a copy
nmax=13
lim=[1,nmax]
img = Image.new('1', (2560+w, 140),255)
draw = ImageDraw.Draw(img)
for n in range(1,nmax+1):
 s = str(n)
 w2, h = draw.textsize(s, font=myFont)
 i=int((w - w2) * .5 + round((n-lim[0]) * 2540. / (lim[1]-lim[0])))
 draw.text((i, 0),s,0,font=myFont)
 if n==1 or n==nmax: continue
 i=int(round((n-lim[0]) * 254. / (lim[1]-lim[0])))
 paletteScaleTicks2[i]=(0,0,0)

img.save('paletteNumbers2.png')
img = Image.new('1', (wpc, 1066), 255)
h=1066-hpc
draw = ImageDraw.Draw(img)
draw.text((0, 0),'0%',0,font=myFont)
draw.text((0, h/3),'100%',0,font=myFont)
draw.text((0, h*2/3),'200%',0,font=myFont)
draw.text((0, h),'300%',0,font=myFont)
img.save('palettePercentages.png')
img = Image.new('1', (wpc, 1066), 255)
h=1066-hpc
draw = ImageDraw.Draw(img)
draw.text((0, 0),'0%',0,font=myFont)
draw.text((0, h/3),'50%',0,font=myFont)
draw.text((0, h*2/3),'100%',0,font=myFont)
draw.text((0, h),'150%',0,font=myFont)
img.save('palettePercentages2.png')
rgbScale2=np.zeros((102,256,3))
rgbScale2[:-1,1:-1,:]=rgbScale
with open('palette.png','wb') as f:
  palettePng.write(f,np.array([paletteScaleTicks,]*3 +
    [paletteScale,]*9).reshape(12,254*3))

with open('palette2.png','wb') as f:
  palettePng.write(f,np.array([paletteScaleTicks,]*3 +
    [paletteScale,]*6 + [paletteScaleTicks2,]*3).reshape(12,254*3))

with open('satPalette2.png','wb') as f:
  satPalette2Png.write(f,rgbScale2.reshape(102,256*3))

#raise ValueError()
header='line,sample,satAge,dayNo,latM,icsenz,airmass,lnChlor_a,Rrs412,Rrs469,Rrs531,Rrs547,Rrs555,Rrs645'.split(',')
header2='line,sample,bias,sd,se'.split(',')
lnChlPos=header.index('lnChlor_a')
ndata=len(header)
flags2={}
for term in header:
   if term[:4]=='flag':
      flags2[term]=2**(int(term[4:])-1)

chunk=1000000
robjects.r('''
  library(gamlss)
  library(geosphere)
  library(Hmisc)
  load(file="scratch/model.RData")
  timediff<--stats$mean['timediff.min']/stats$sd['timediff.min']
  sdMin<-min(predict(model,what="sigma",type="response"))
''')
meanSds = {}
with open('meanSds.csv') as f:
  reader = csv.reader(f)
  temp = next(reader)
  for line in reader:
    meanSds[line[0]] = [float(i) for i in line[1:]]

rgbPngMap = png.Writer(nsc,nlc)
land = np.zeros((nlc, nsc), dtype = bool)
inFiles = sorted(glob.glob('1-8May2013/A*/*.L2_LAC'))
nFiles = len(inFiles)
if processing:
 for inFile in inFiles:
  tail = inFile.split('/')[-1]
  root = inFile[:-7]
  print(tail)
  dayNo = int(tail[5:8])
  myDate = date(int(tail[1:5]),1,1) + timedelta(dayNo - 1)
  satAge = (myDate - launchDate).days
  modisData={'satAge':satAge, 'dayNo':dayNo}
  with Dataset(inFile) as nc:
    nl = nc.__getattr__('Number of Scan Lines')
    ns = nc.__getattr__('Pixels per Scan Line')
    # 0=lnChl,1=weight,2=lnChl2,3=weight2,4=land,5=passPresence,6=lnChlMask
    reprojData = np.ma.empty((nl, ns, 10))
    ncv=nc.variables
    var=ncv['chlor_a']
    chl=var[::-1,::-1]
    chl=np.ma.masked_array(chl,mask=chl<-32766) # masked
    #chl=chl*var.slope+var.intercept
    reprojData[:,:,0]=np.ma.log(chl)
    lnChl=reprojData[:,:,0]
    reprojData[:,:,6]=chl.mask
    modisData['lnChlor_a']=lnChl
    var=ncv['latitude']
    modisData['latM']=var[::-1,::-1]
    var=ncv['longitude']
    modisData['lonM']=var[::-1,::-1]
    loadModisData(['Rrs412','Rrs469','Rrs531','Rrs547','Rrs555','Rrs645',
      'senz','solz'])
    modisData['icsenz']=1./np.cos(np.radians(modisData['senz']))
    modisData['senz']=0.
    modisData['airmass']=(modisData['icsenz'] +
      1./np.cos(np.radians(modisData['solz'])))
    modisData['solz']=0.
    l2flags=ncv['l2_flags'][::-1,::-1]
  for i in meanSds:
    if i not in ['windspeed', 'dayLength', 'Rrs443', 'Rrs488', 'Rrs667',
        'Rrs678', 'icsolz', 'timediff.min']:
      modisData[i] = (modisData[i] - meanSds[i][0]) / meanSds[i][1]
  flags = {}
  for i in range(32):
    flags['flag' + str(i + 1)] = (l2flags & 2**i) != 0 # L3
  swathLand = flags['flag2']
  reprojData[:,:,4] = swathLand
  reprojData[:,:,5] = 1 # pass presence flag
  ml, ms = nl - 1, ns - 1
  bias = np.ma.masked_all((nl, ns))
  sd = np.ma.masked_all((nl, ns))
  se = np.ma.masked_all((nl, ns))
  ctPng=png.Writer(ns,nl,palette=chlScale)
  rgbPng=png.Writer(ns,nl)
  chl8bit=writeChl(chl,root + 'chl.png')
  swath_def = geometry.SwathDefinition(lons=modisData['lonM'],
    lats=modisData['latM'])
  data=['0']*ndata
  tile=0
  newFile=True
  line = -1
  sample = ns
  finished = False
  while True:
    if sample == ns:
      line += 1
      #if line % 500 == 0: print(line, nl)
      if line == nl:
        finished = True
      else:
        sample = 0
    if (not finished) and chl.mask[line, sample]:
      sample += 1
      continue
    if finished or newFile:
      if finished or tile != 0: # run predict in R
        f.close()
        robjects.r('''
          d<-read.csv("''' + csvFile + '''")
          line<-d$line
          sample<-d$sample
          latM<-d$latM
          dayNo<-d$dayNo*stats$sd['dayNo']+stats$mean['dayNo']
          d<-d[,!names(d) %in% c("line", "sample", "latM")]
          d$dayLength=(daylength(latM,dayNo)-stats$mean['dayLength'])/
            stats$sd['dayLength']
          d$timediff.min=timediff
          #print(d[1,])
          #print('predicting bias')
          bias<-predict(model,newdata=d,type="response")
          #print(summary(bias))
          #print('predicting sd')
          sd<-predict(model,what="sigma",newdata=d,type="response")
          sd[sd<sdMin]<-sdMin
          #print(summary(sd))
          #print('predicting se')
          se<-0
          # predict doesn't do se, so do it ourselves.
          # Need an assumption about variable correlation,
          # initially assume uncorrelated.
          pf<-environment(formula(model))
          data <- eval(model$call$data, pf)
          t <- lpred(model,type="terms",se.fit=TRUE)
          nt<-ncol(tms<-as.matrix(t$fit))
          cn<-parse(text=colnames(tms))
          for (i in 1:nt){
            cni<-cn[[i]]
            if (length(cni)>1) cni<-cni[[2]]
            xx<-eval(cni,data,pf)
            yy<-t$se.fit[,i]
            name<-toString(cni)
            se<-se+approxExtrap(xx, yy, d[,name])$y^2
            #print(paste(name,i,nt,length(se),length(xx),length(yy),length(d[,name])))
          }
          se<-sqrt(se)
          #print(summary(se))
          # predict converts from log already, no need to do it again
          write.table(data.frame(line=line, sample=sample, bias=bias,
            sd=sd, se=se), file="''' + csvFile2 + '''",
            quote=FALSE, row.names=FALSE, sep=",")
        ''')
      if finished: break
      csvFile = inFile + str(tile) + '.csv'
      csvFile2 = inFile + str(tile) + 'V2.csv'
      #print(csvFile)
      f = open(csvFile, 'wb')
      writer = csv.writer(f)
      writer.writerow(header)
      newFile = False
      count = 0
    data[:4] = [str(line), str(sample), str(modisData['satAge']),
      str(modisData['dayNo'])]
    l2f = l2flags[line, sample]
    for i in range(4, ndata):
      if header[i][:4] == 'flag':
        data[i] = str(((l2f & flags2[header[i]]) > 0) * 1)
      elif np.ma.is_masked(modisData[header[i]][line, sample]):
        data[i] = 'NA'
        if header[i][:3]=='Rrs':
          print('Bad Rrs data',inFile,i,header[i],line,sample)
        else:
          raise ValueError('Bad data',inFile,i,header[i],line,sample)
      else:
        data[i] = str(modisData[header[i]][line, sample])
    if 'NA' not in data:
      writer.writerow(data)
    sample += 1
    count += 1
    if count == chunk:
      newFile = True
      tile += 1
  for csvFile in glob.glob(inFile + '*V2.csv'):
    with open(csvFile) as f:
      reader = csv.reader(f)
      temp = next(reader)
      for data in reader:
        line, sample = (int(data[0]), int(data[1]))
        if chl.mask[line, sample]:
          raise ValueError(line,sample)
        bias[line, sample] = float(data[2])
        sd[line, sample] = float(data[3])
        se[line, sample] = float(data[4])
  error = np.ma.concatenate((bias, sd, se))
  eMax = np.percentile(error[~error.mask], 95)
  print('eMax', eMax)
  error = bias**2 + sd**2 + se**2
  rgb = np.empty((nl, ns, 3), int)
  x = np.clip((abs(bias) * 256. / eMax).astype(int), 0, 255)
  x[error.mask] = 255
  x[swathLand] = 100
  rgb[..., 0] = x
  x = np.clip((sd * 256. / eMax).astype(int), 0, 255)
  x[error.mask] = 255
  x[swathLand] = 100
  rgb[..., 1] = x
  x = np.clip((se * 256. / eMax).astype(int), 0, 255)
  x[error.mask] = 255
  x[swathLand] = 100
  rgb[..., 2] = x
  with open(root + 'errorRgb.png','wb') as f:
   rgbPng.write(f,rgb.reshape((nl,ns*3)))
  reprojData[:,:,1] = 1 / error
  reprojData[:,:,7] = bias
  reprojData[:,:,8] = sd
  reprojData[:,:,9] = se
  error = error**.5
  errorNoMask=error[~error.mask]
  #print(np.ma.size(error),len(errorNoMask),errorNoMask.min(),errorNoMask.max())
  print([5,25,50,75,95])
  print(np.percentile(errorNoMask,[5,25,50,75,95]))
  p=plt.hist(errorNoMask,bins=np.arange(100.)*.05)
  L3mask = lnChl.mask.copy()
  for i in l3flags:
    L3mask |= flags['flag' + str(i)]
  error2=error.copy()
  error2.mask |= ~L3mask # keep only valid chl masked at L3
  p=plt.hist(error2[~error2.mask],bins=np.arange(100.)*.05)
  plt.savefig(root + 'errorHistogram.png')
  plt.close()
  p1=500.
  rgb = writeSatChl(chl, error, root + 'satChlError500.png')
  chlBias8bit=writeChl(chl*(np.abs(np.ma.exp(bias)-1)),root+'chlBias.png')
  chlSd8bit=writeChl(chl*(np.ma.exp(sd)-1), root + 'chlSd.png')
  chlSe8bit=writeChl(chl*(np.ma.exp(se)-1), root + 'chlSe.png')
  chlError8bit=writeChl(chl*(np.ma.exp(error)-1), root + 'chlError.png')
  reprojData[:,:,2]=lnChl.copy()
  lnChl2=reprojData[:,:,2]
  lnChl2[~bias.mask]-=bias[~bias.mask]
  lnChl2.mask|=bias.mask
  chl2=np.ma.exp(lnChl2)
  chl28bit=writeChl(chl2, root + 'chl2.png')
  residual = sd**2 + se**2
  reprojData[:,:,3] = 1. / residual
  residual = residual**.5
  chlResidual=chl2*np.ma.exp(residual)
  chlResidual8bit=writeChl(chlResidual, root + 'chlResidual.png')
  rgb=writeSatChl(chl2,residual, root + 'satChlResidual500.png')
  p1=300.
  rgb=writeSatChl(chl,error, root + 'satChlError300.png')
  rgb=writeSatChl(chl2,residual, root + 'satChlResidual300.png')
  lnChl = reprojData[:,:,0].copy()
  for i in range(4):
    reprojData.mask[:,:,i] = L3mask
  result = kd_tree.resample_nearest(swath_def, reprojData, area_def,
    radius_of_influence=10000, fill_value = None)
  rgb = np.empty((nlc, nsc, 3), int)
  eMax = 1.
  biasc = result[..., 7]
  sdc = result[..., 8]
  sec = result[..., 9]
  errorc = biasc**2 + sdc**2 + sec**2
  x = np.clip((abs(biasc) * 256. / eMax), 0, 255).astype(int)
  x[errorc.mask] = 255
  rgb[..., 0] = x
  x = np.clip(sdc * 256. / eMax, 0, 255).astype(int)
  x[errorc.mask] = 255
  rgb[..., 1] = x
  x = np.clip(sec * 256. / eMax, 0, 255).astype(int)
  x[errorc.mask] = 255
  rgb[..., 2] = x
  with open(root + 'errorRgbMap.png','wb') as f:
   rgbPngMap.write(f, rgb.reshape((nlc, nsc * 3)))
  with Dataset(inFile + '_mapped.nc', 'w') as nc:
    nc.createDimension('latitude',nlc)
    nc.createDimension('longitude',nsc)
    nc.createVariable('latitude','f',('latitude'))
    nc.createVariable('longitude','f',('longitude'))
    nc.variables['latitude'][:] = np.linspace(limits[1], limits[0], nlc)
    nc.variables['longitude'][:] = np.linspace(limits[2], limits[3], nsc)
    l3vi = nc.createVariable('L3lnChl', 'f', ('latitude', 'longitude'),
      fill_value = fill)
    l3vi[:] = result[:,:,0]
    l3vi = nc.createVariable('lnChlWeight', 'f', ('latitude', 'longitude'),
      fill_value = fill)
    l3vi[:] = result[:,:,1]
    l3vi = nc.createVariable('L3lnChl2', 'f', ('latitude', 'longitude'),
      fill_value = fill)
    l3vi[:] = result[:,:,2]
    l3vi = nc.createVariable('lnChl2Weight', 'f', ('latitude', 'longitude'),
      fill_value = fill)
    l3vi[:] = result[:,:,3]
    l3vi = nc.createVariable('land', 'i', ('latitude', 'longitude'),
      fill_value = -1)
    l3vi[:] = result[:,:,4]
    l3vi = nc.createVariable('bias', 'f', ('latitude', 'longitude'),
      fill_value = fill)
    l3vi[:] = result[:,:,7]
    l3vi = nc.createVariable('sd', 'f', ('latitude', 'longitude'),
      fill_value = fill)
    l3vi[:] = result[:,:,8]
    l3vi = nc.createVariable('se', 'f', ('latitude', 'longitude'),
      fill_value = fill)
    l3vi[:] = result[:,:,9]

  l = result[:,:,4]
  l._sharedmask=False # use future behaviour of shared masks https://stackoverflow.com/questions/41028253/numpy-1-13-maskedarrayfuturewarning-setting-an-item-on-a-masked-array-which-has
  l[l.mask] = 0
  land[l != 0] = True
  if land.min() != False:
    print(inFile,land.min(),l.min(),l.mask.max(),np.count_nonzero(l != 0))
  cLnChl = result[:,:,0]
  w = result[:,:,1]
  ctPng=png.Writer(nsc,nlc,palette=chlScale)
  cChl8bit=writeChl(cLnChl,root + 'chlMap.png',l,1)
  cChlError8bit=writeChl(cLnChl+np.ma.log(np.ma.exp(w**(-.5))-1),
    root + 'chlMapError.png',l,1)
  cLnChl = result[:,:,2]
  w = result[:,:,3]
  cChl8bit=writeChl(cLnChl,root + 'chl2Map.png',l,1)
  cChlResidual8bit=writeChl(cLnChl+np.ma.log(np.ma.exp(w**(-.5))-1),
    root + 'chlMapResidual.png',l,1)
  nc2=Dataset(inFile+'_chlerrors.nc','w')
  nc2.createDimension('line',nl)
  nc2.createDimension('sample',ns)
  nc2.createVariable('latitude','f',('line','sample'),fill_value=fill)
  nc2.variables['latitude'][:]=modisData['latM']
  nc2.createVariable('longitude','f',('line','sample'),fill_value=fill)
  nc2.variables['longitude'][:]=modisData['lonM']
  nc2.createVariable('l2flags','int32',('line','sample'))
  nc2.variables['l2flags'][:]=l2flags
  nc2.createVariable('chl','f',('line','sample'),fill_value=fill)
  nc2.variables['chl'][:]=chl
  nc2.createVariable('lnchl','f',('line','sample'),fill_value=fill)
  nc2.variables['lnchl'][:]=lnChl
  nc2.createVariable('lnchlL3masked','f',('line','sample'),fill_value=fill)
  nc2.variables['lnchlL3masked'][:]=reprojData[:,:,0]
  nc2.createVariable('chl2','f',('line','sample'),fill_value=fill)
  nc2.variables['chl2'][:]=chl2
  nc2.createVariable('lnchl2','f',('line','sample'),fill_value=fill)
  nc2.variables['lnchl2'][:]=lnChl2
  nc2.createVariable('lnchl2L3masked','f',('line','sample'),fill_value=fill)
  nc2.variables['lnchl2L3masked'][:]=reprojData[:,:,2]
  nc2.createVariable('bias','f',('line','sample'),fill_value=fill)
  nc2.variables['bias'][:]=bias
  nc2.createVariable('sd','f',('line','sample'),fill_value=fill)
  nc2.variables['sd'][:]=sd
  nc2.createVariable('se','f',('line','sample'),fill_value=fill)
  nc2.variables['se'][:]=se
  nc2.createVariable('error','f',('line','sample'),fill_value=fill)
  nc2.variables['error'][:]=error
  nc2.createVariable('residual','f',('line','sample'),fill_value=fill)
  nc2.variables['residual'][:]=residual
  nc2.createVariable('chlErrorPercent','f',('line','sample'),
    fill_value=fill)
  nc2.variables['chlErrorPercent'][:]=(np.ma.exp(error)-1.)*100.
  nc2.createVariable('chlResidualPercent','f',('line','sample'),
    fill_value=fill)
  nc2.variables['chlResidualPercent'][:]=(np.ma.exp(residual)-1.)*100.
  nc2.close()
 with open('land.png', 'wb') as f:
  w = png.Writer(nsc, nlc, greyscale = True)
  w.write(f,land)
else:
 r = png.Reader('land.png')
 d = r.read()
 land =  np.vstack(itertools.imap(np.uint8, d[2]))
 land.dtype = bool
ncs = [] # list of netCDF file handles
for inFile in inFiles:
  ncs += [Dataset(inFile + '_mapped.nc')]

nc=Dataset('1-8May2013/mappedComposite.nc','w')
nc.createDimension('latitude',nlc)
nc.createDimension('longitude',nsc)
nc.createVariable('latitude','f',('latitude'))
nc.createVariable('longitude','f',('longitude'))
nc.variables['latitude'][:] = np.linspace(limits[1], limits[0], nlc)
nc.variables['longitude'][:] = np.linspace(limits[2], limits[3], nsc)
l3nv = nc.createVariable('n', 'f', ('latitude', 'longitude'),
  fill_value = fill)
l3v = nc.createVariable('L3lnChl', 'f', ('latitude', 'longitude'),
  fill_value = fill)
wl3v = nc.createVariable('L3wLnChl', 'f', ('latitude', 'longitude'),
  fill_value = fill)
wl3uv = nc.createVariable('L3wLnChlUncertainty', 'f',
  ('latitude', 'longitude'), fill_value = fill)
wl3cv = nc.createVariable('L3wLnChl2', 'f', ('latitude', 'longitude'),
  fill_value = fill)
wl3csv = nc.createVariable('L3wLnChl2ExcessSigma', 'f',
  ('latitude', 'longitude'), fill_value = fill)
wl3cuv = nc.createVariable('L3wLnChl2Uncertainty', 'f',
  ('latitude', 'longitude'), fill_value = fill)
for line in xrange(nlc): # read files line by line to calculate stats
  if line % 100 == 0: print(line, nlc)
  lineData = [[], [], [], []] # lnChl, weight, lnChl2, weight2
  for nc2 in ncs:
    x = nc2.variables['L3lnChl'][line, :]
    if type(x) != np.ma.MaskedArray:
      if x.min() > -1e36:
        raise ValueError(x.min())
      x = np.ma.array(x, mask = x < -1e36)
    lineData[0] += [x]
    lineData[1] += [nc2.variables['lnChlWeight'][line, :]]
    lineData[2] += [nc2.variables['L3lnChl2'][line, :]]
    lineData[3] += [nc2.variables['lnChl2Weight'][line, :]]
  lnChl = np.ma.array(lineData[0])
  weight = np.ma.array(lineData[1])
  lnChl2 = np.ma.array(lineData[2])
  weight2 = np.ma.array(lineData[3])
  if (weight.mask != lnChl.mask).any(): raise ValueError(line)
  if (lnChl2.mask != lnChl.mask).any(): raise ValueError(line)
  if (weight2.mask != lnChl.mask).any(): raise ValueError(line)
  cw = np.ma.masked_all(nsc)
  cw2 = np.ma.masked_all(nsc)
  excessVar = np.ma.masked_all(nsc)
  n = (~lnChl.mask).sum(axis = 0)
  multi = np.nonzero(n > 1)[0] # consider scatter for n>1
  l3nv[line, :] = n
  l3v[line, :] = lnChl.sum(axis = 0) / n
  wmean2 = (lnChl2 * weight2).sum(axis = 0) / weight2.sum(axis = 0)
  dlnChl2 = lnChl2.copy()
  for x in dlnChl2: x -= wmean2 # subtract weighted mean from corrected lnChl
  res2 = dlnChl2**2 * weight2 # residual^2 = (lnChl-bias-wmean)^2/(SD^2+SE^2)
  sigmar2 = res2[:, multi].sum(axis = 0) / (n[multi] - 1)
  excess = np.nonzero(sigmar2 > 1.)[0] # account for excess sigma as in-pixel variability
  #print('s2',sigmar2[:29])
  #print('e',excess)
  excessVar[multi] = 0. # zero if n>1 but no excess
  excessVar.mask[multi] = False
  for pos in excess:
    mp = multi[pos]
    nmp = n[mp]
    lc = lnChl2[:, mp]
    good = ~lc.mask
    lc = lc[good]
    wm = wmean2[mp]
    E = sigmar2[pos] - 1.
    w = weight2[good, mp]
    iw0 = w**(-1)
    sn2 = 0.
    #print('mp',mp,nmp,lc,sd2se2)
    for i in xrange(1000):
      oldAbsE = abs(E)
      sn2 += E / ((w * (lc - wm))**2).sum()
      #print(i,wm,dlc2,E,sn2)
      w = 1. / (iw0 + sn2)
      wm = (w * lc).sum() / w.sum()
      E = (w * (lc - wm)**2).sum() / (nmp - 1) - 1.
      absE = abs(E)
      if absE < 1.2e-6: break # chosen by trial and error to avoid rounding issues
      if absE > oldAbsE: raise ValueError(pos,mp,nmp,lc,i,wm,E,oldAbsE)
    if absE >= 1.2e-6: raise ValueError(pos,mp,nmp,lc,i,wm,E,oldAbsE)
    #print('done', i, sc2, wm, dlc2, sr2, oldSr2)
    wmean2[mp] = wm
    excessVar[mp] = sn2
    weight[good, mp] = 1. / (1. / weight[good, mp] + sn2)
    weight2[good, mp] = w
    if (weight2[good, mp] < weight[good, mp]).any():
      raise ValueError(pos, mp, weight2[good, mp], weight[good, mp])
  wl3cv[line, :] = wmean2
  wl3csv[line, :] = excessVar**.5
  wl3uv[line, :] = weight.sum(axis = 0)**(-.5)
  wl3cuv[line, :] = weight2.sum(axis = 0)**(-.5)
  l3v[line, :] = lnChl.sum(axis = 0) / n
  wl3v[line, :] = (lnChl * weight).sum(axis = 0) / weight.sum(axis = 0)
  nc.sync()

ctPng=png.Writer(nsc,nlc,palette=chlScale)
n=l3nv[:]
n8bit=writeChl(n,'1-8May2013/compositeNmap.png',land,0,[1,n.max()]) # 7a
cChl8bit=writeChl(l3v[:],'1-8May2013/compositeChlMap.png',land,1) # 7b
chl=np.ma.exp(wl3v[:])
cChl8bit=writeChl(chl,'1-8May2013/compositeWeightedChlMap.png',land) # 8a
u=wl3uv[:]
print([1,5,25,50,75,95,99])
print(np.percentile((np.ma.exp(u[~u.mask])-1.)*100.,[1,5,25,50,75,95,99]))
cChl8bit=writeChl(chl*(np.ma.exp(u)-1),'1-8May2013/compositeWeightedChlUncertaintyMap.png',land) # 8b
rgbPng=rgbPngMap
p1=300.
rgb=writeSatChl(chl,u, '1-8May2013/compositeSatChlError300.png',land) # 9a
p1=150.
rgb=writeSatChl(chl,u, '1-8May2013/compositeSatChlError150.png',land) # 9a
chl=np.ma.exp(wl3cv[:])
cChl8bit=writeChl(chl,'1-8May2013/compositeWeightedChl2Map.png',land) # 8c
u=wl3cuv[:]
print(np.percentile((np.ma.exp(u[~u.mask])-1.)*100.,[1,5,25,50,75,95,99]))
cChl8bit=writeChl(chl*(np.ma.exp(u)-1),'1-8May2013/compositeWeightedChl2UncertaintyMap.png',land) # 8d
p1=300.
rgb=writeSatChl(chl,u, '1-8May2013/compositeSatChlResidual300.png',land) # 9b
p1=150.
rgb=writeSatChl(chl,u, '1-8May2013/compositeSatChlResidual150.png',land) # 9b
chlScale[1]=(150,150,150) # make zero grey
ctPng=png.Writer(nsc,nlc,palette=chlScale)
u=wl3csv[:]
print(u.min(),u.max())
u=chl*(np.ma.exp(u)-1)
chlDn2=10**(2.*.015-2.)
u[np.logical_and(u>0,u<chlDn2)]=chlDn2
cChl8bit=writeChl(u,'1-8May2013/compositeWeightedChl2ExcessSdMap.png',land)
nc.close()
