from __future__ import print_function
import numpy as np

def readCoeffs(fileName, ranges):

   with open(fileName) as f:
      terms = {'linkFunction':f.readline().split()[-1],
         'parameter':f.readline().split()[0]}
      titles = f.readline()
      for row in f:
         if row == '' or row == '\n':
            break
         data = row.split()
         try:
            name = data[0]
         except:
            raise NameError('Odd row', row, '.')
         if name == '(Intercept)':
            terms['intercept'] = float(data[1])
            continue
         if name[:-1] in terms:
            try:
               power = int(name[-1:])
            except:
               raise NameError('Power expected in', name)
            name = name[:-1]
            if isinstance(terms[name], np.ndarray):
               if len(terms[name]) > power:
                  raise NameError('Unusual power', power, terms[name])
            else:
               if power < 2:
                  raise NameError('Power 2+ expected', power, terms[name])
               terms[name] = np.array([terms[name], 0.])
            while len(terms[name]) < power:
               terms[name] = np.insert(terms[name], 0, 0.)
            terms[name] = np.insert(terms[name], 0, float(data[1]))
         else:
            terms[name] = float(data[1])

   for term in terms:
      if isinstance(terms[term], np.ndarray):
         if not term in ranges:
            raise NameError(term, 'not found in ranges')
         coeffs = np.copy(terms[term])
         grad = np.polyder(coeffs)
         for i in range(2):
            x = ranges[term][i]
            y = np.polyval(coeffs, x)
            dy = np.polyval(grad, x)
            if i == 0:
               terms[term] = [np.array([dy, y - dy * x]), coeffs]
            else:
               terms[term] += [np.array([dy, y - dy * x])]
         terms[term] = {'range':ranges[term], 'coeffs':terms[term]}

   return terms

